
import logging
from functools import lru_cache
from typing import List, Optional

from brownie import chain
from y import convert
from y.contracts import has_method, has_methods
from y.datatypes import UsdPrice, UsdValue
from y.decorators import log
from y.erc20 import decimals
from y.networks import Network
from y.prices import magic
from y.typing import Address, AddressOrContract, AnyAddressType, Block
from y.utils.multicall import \
    multicall_same_func_same_contract_different_inputs
from y.utils.raw_calls import _totalSupplyReadable

logger = logging.getLogger(__name__)


@lru_cache
@log(logger)
def is_saddle_lp(token_address: AnyAddressType) -> bool:
    pool = get_pool(token_address)
    if pool: return has_methods(pool, ['getVirtualPrice()(uint)', 'getA()(uint)','getAPrecise()(uint)'])


@lru_cache
@log(logger)
def get_pool(token_address: AnyAddressType) -> Address:
    convert.to_address(token_address)
    if chain.id == Network.Mainnet:
        if token_address == '0xc9da65931ABf0Ed1b74Ce5ad8c041C4220940368': # saddle aleth doesn't have swap() function
            return '0xa6018520EAACC06C30fF2e1B3ee2c7c22e64196a'
        elif token_address == '0xd48cF4D7FB0824CC8bAe055dF3092584d0a1726A': # saddle d4
            return '0xC69DDcd4DFeF25D8a793241834d4cc4b3668EAD6'
        elif token_address == '0xF32E91464ca18fc156aB97a697D6f8ae66Cd21a3':
            return '0xdf3309771d2BF82cb2B6C56F9f5365C8bD97c4f2'
    pool = has_method(token_address, 'swap()(address)', return_response=True)
    return pool or None


@log(logger)
def get_price(token_address: AddressOrContract, block: Optional[Block] = None) -> UsdPrice:
    return UsdPrice(tvl(token_address, block) / _totalSupplyReadable(token_address, block))


@log(logger)
def tvl(token_address: AnyAddressType, block: Optional[Block] = None) -> UsdValue:
    tokens = get_tokens(token_address, block)
    pool = get_pool(token_address)
    balances = multicall_same_func_same_contract_different_inputs(
        pool, 'getTokenBalance(uint8)(uint)', inputs=[*range(len(tokens))])
    tokens_decimals = decimals(tokens, block=block)
    balances = [balance / 10 ** decimal for balance, decimal in zip(balances, tokens_decimals)]
    prices = magic.get_prices(tokens, block, silent=True)
    return UsdValue(sum(balance * price for balance, price in zip (balances, prices)))


@log(logger)
def get_tokens(token_address: AnyAddressType, block: Optional[Block] = None) -> List[Address]:
    pool = get_pool(token_address)
    response = multicall_same_func_same_contract_different_inputs(
        pool, 'getToken(uint8)(address)', inputs=[*range(8)], block=block, return_None_on_failure=True)
    return [token for token in response if token is not None]




from typing import Optional

from brownie import chain
from brownie.convert.datatypes import EthAddress
from y.datatypes import UsdPrice
from y.networks import Network
from y.prices import magic
from y.typing import Block

MAPPING = {
    Network.Mainnet: {
        "0x4da27a545c0c5B758a6BA100e3a049001de870f5": "0x7Fc66500c84A76Ad7e9c93437bFc5Ac33E2DDaE9", # stkaave -> aave
    }
}.get(chain.id,{})

def is_one_to_one_token(token_address: EthAddress) -> bool:
    return token_address in MAPPING

def get_price(token_address: EthAddress, block: Optional[Block] = None) -> UsdPrice:
    return magic.get_price(MAPPING[token_address], block=block)



from functools import lru_cache
from typing import Optional, Tuple

from brownie.convert.datatypes import EthAddress
from brownie.exceptions import ContractNotFound
from y import Contract
from y.classes.common import ERC20, WeiBalance
from y.datatypes import UsdPrice, UsdValue
from y.exceptions import ContractNotVerified, MessedUpBrownieContract
from y.typing import AnyAddressType, Block
from y.utils.cache import memory
from y.utils.multicall import fetch_multicall


@memory.cache()
def is_generic_amm(lp_token_address: AnyAddressType) -> bool:
    try:
        token_contract = Contract(lp_token_address)
        return all(hasattr(token_contract, attr) for attr in ['getReserves','token0','token1'])
    except (ContractNotFound, ContractNotVerified):
        return False
    except MessedUpBrownieContract:
        # probably false
        return False
        
class GenericAmm:
    def __contains__(self, lp_token_address: AnyAddressType) -> bool:
        return is_generic_amm(lp_token_address)
    
    def get_price(self, lp_token_address: AnyAddressType, block: Optional[Block] = None) -> UsdPrice:
        lp_token_contract = Contract(lp_token_address)
        total_supply, decimals = fetch_multicall(*[[lp_token_contract, attr] for attr in ['totalSupply','decimals']], block=block)
        total_supply_readable = total_supply / 10 ** decimals
        return UsdPrice(self.get_tvl(lp_token_address, block) / total_supply_readable)

    @lru_cache(maxsize=None)
    def get_tokens(self, lp_token_address: AnyAddressType) -> Tuple[ERC20,ERC20]:
        lp_token_contract = Contract(lp_token_address)
        tokens = fetch_multicall(*[[lp_token_contract,attr] for attr in ['token0', 'token1']])
        return ERC20(tokens[0]), ERC20(tokens[1])
    
    def get_tvl(self, lp_token_address: AnyAddressType, block: Optional[Block] = None) -> UsdValue:
        lp_token_contract = Contract(lp_token_address)
        tokens = self.get_tokens(lp_token_address)
        reserves = lp_token_contract.getReserves(block_identifier=block)
        reserves = [WeiBalance(reserve,token,block) for token, reserve in zip(tokens,reserves)]
        return UsdValue(sum(reserve.value_usd() for reserve in reserves))


generic_amm = GenericAmm()
